from __future__ import annotations

import ckan.plugins.toolkit as tk


SELFTOOLS_CATEGORIES = "ckan.selftools.categories"
SELFTOOLS_OPERATIONS_PWD = "ckan.selftools.opetations_pwd"
SELFTOOLS_OPERATIONS_LIMIT = "ckan.selftools.operations_limit"
SELFTOOLS_CONFIG_BLACKLIST = "ckan.selftools.config_blacklist"
SELFTOOLS_TOOLS_BLACKLIST = "ckan.selftools.tools_blacklist"
SELFTOOLS_MODEL_FIELDS_BLACKLIST = "ckan.selftools.model_fields_blacklist"
SELFTOOLS_MODEL_ENCRYPTION_KEY = "ckan.selftools.model_key_ecryption"


def selftools_get_categories_list():
    return tk.config.get(SELFTOOLS_CATEGORIES)


def selftools_get_operations_pwd():
    return tk.config.get(SELFTOOLS_OPERATIONS_PWD)


def selftools_get_operations_limit():
    return tk.config.get(SELFTOOLS_OPERATIONS_LIMIT)


def selftools_get_config_blacklist():
    return tk.config.get(SELFTOOLS_CONFIG_BLACKLIST)


def selftools_get_tools_blacklist():
    return tk.config.get(SELFTOOLS_TOOLS_BLACKLIST)


def selftools_get_model_fields_blacklist():
    return tk.config.get(SELFTOOLS_MODEL_FIELDS_BLACKLIST)


def selftools_get_model_ecryption_key():
    return tk.config.get(SELFTOOLS_MODEL_ENCRYPTION_KEY)



from __future__ import annotations

from typing import Any

from ckanext.selftools import utils, config


def selftools_categories() -> list[Any]:
    return utils.get_selftools_categories()


def selftools_get_db_model_options() -> list[dict[str, Any]]:
    models: list[dict[str, Any]] = utils.get_db_models()
    return [{"value": i.get("label"), "text": i.get("label")} for i in models]


def get_operations_limit() -> int:
    return config.selftools_get_operations_limit()


def check_operations_pwd_set() -> bool:
    return True if config.selftools_get_operations_pwd() else False



from __future__ import annotations

from typing import Any

from ckan.plugins import Interface


class ISelftools(Interface):
    """Implement custom Selftools response modification."""

    def selftools_db_models(self, models_list: list[Any]) -> list[Any]:
        """Return DB models list.
        This method is called after the models list is prepared and before it is returned.
        Useful if you want to add additional models or custom models.
        :returns: list
        :rtype: list

        """

        return models_list



from __future__ import annotations

import json
from typing import cast, Any
from flask import Blueprint, Response
from urllib.parse import urlencode
from cryptography.fernet import Fernet, InvalidToken
import gzip
from werkzeug.utils import secure_filename

from ckan import types
import ckan.lib.navl.dictization_functions as dict_fns
import ckan.logic as logic
import ckan.model as model
import ckan.plugins.toolkit as tk
from ckan.common import _, request

import ckanext.selftools.config as selftools_config


selftools_htmx = Blueprint("selftools_htmx", __name__)


@selftools_htmx.route("/selftools/solr-query", methods=["POST"])
def selftools_solr_query() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_solr_query")(context, data_dict)

    pretty_json = json.dumps(resp, indent=2)

    return tk.render(
        "/selftools/results/pretty_json.html", extra_vars={"json": pretty_json}
    )


@selftools_htmx.route("/selftools/solr-delete", methods=["POST"])
def selftools_solr_delete() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_solr_delete")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Couldn't delete index.")
    else:
        return _("Deleted.")


@selftools_htmx.route("/selftools/solr-index", methods=["POST"])
def selftools_solr_index() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_solr_index")(context, data_dict)

    if not resp.get("success"):
        return (
            resp["message"]
            if resp.get("message")
            else _("Couldn't index dataset. No such Dataset.")
        )
    else:
        return _("Indexed.")


@selftools_htmx.route("/selftools/db-query", methods=["POST"])
def selftools_db_query() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_db_query")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Something went wrong...")
    else:
        return tk.render(
            "/selftools/results/db_results.html", extra_vars={"data": resp}
        )


@selftools_htmx.route("/selftools/db-update", methods=["POST"])
def selftools_db_update() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_db_update")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Something went wrong...")
    else:
        return tk.render(
            "/selftools/results/db_effected.html", extra_vars={"data": resp}
        )


@selftools_htmx.route("/selftools/redis-query", methods=["POST"])
def selftools_redis_query() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_redis_query")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Something went wrong...")
    else:
        return tk.render(
            "/selftools/results/redis_results.html", extra_vars={"data": resp}
        )


@selftools_htmx.route("/selftools/redis-update", methods=["POST"])
def selftools_redis_update() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_redis_update")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Something went wrong...")
    else:
        return _("Updated/Created.")


@selftools_htmx.route("/selftools/redis-delete", methods=["POST"])
def selftools_redis_delete() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_redis_delete")(context, data_dict)

    if not resp.get("success"):
        return (
            resp["message"]
            if resp.get("message")
            else _("Couldn't delete Key. No such Key.")
        )
    else:
        return _("Deleted.")


@selftools_htmx.route("/selftools/config-query", methods=["POST"])
def selftools_config_query() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_config_query")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Something went wrong...")
    else:
        return tk.render(
            "/selftools/results/config_results.html", extra_vars={"data": resp}
        )


@selftools_htmx.route("/selftools/model-export", methods=["POST"])
def selftools_model_export() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_model_export")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Something went wrong...")
    else:
        r_length = len(resp["results"])
        show_100 = dict(list(resp["results"].items())[:100])
        pretty_json = json.dumps(show_100, indent=2)

        query_string = urlencode(data_dict, doseq=True)
        download_link = "?".join(
            [
                tk.h.url_for(
                    "selftools_htmx.selftools_model_export_download", _external=True
                ),
                query_string,
            ]
        )

        return tk.render(
            "/selftools/results/model_results.html",
            extra_vars={
                "json": pretty_json,
                "download_link": download_link,
                "r_length": r_length,
            },
        )


@selftools_htmx.route("/selftools/model-export-download", methods=["GET"])
def selftools_model_export_download() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    try:
        data_dict = {k: v if len(v) > 1 else v[0] for k, v in request.args.lists()}

    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    resp = tk.get_action("selftools_model_export")(context, data_dict)

    if not resp.get("success"):
        return resp["message"] if resp.get("message") else _("Something went wrong...")
    else:
        json_bytes = json.dumps(resp["results"]).encode("utf-8")
        compressed = gzip.compress(json_bytes)
        key = selftools_config.selftools_get_model_ecryption_key()

        if key:
            fernet = Fernet(key)
            encrypted = fernet.encrypt(compressed)

            response = Response(encrypted, mimetype="application/octet-stream")
            response.headers["Content-Disposition"] = (
                "attachment; filename=model_export.bin"
            )
        else:
            response = Response(compressed, mimetype="application/gzip")
            response.headers["Content-Disposition"] = (
                "attachment; filename=model_export.gz"
            )

        return response


@selftools_htmx.route(
    "/selftools/model-export-custom-relationships-fields", methods=["POST"]
)
def selftools_model_export_custom_relationships_fields() -> Any | str:
    return tk.render(
        "/selftools/tools/model/model_export_custom_relationships_fields.html",
        extra_vars={},
    )


@selftools_htmx.route(
    "/selftools/model-export-default-value-condition", methods=["POST"]
)
def selftools_model_export_custom_default_value_condition() -> Any | str:
    return tk.render(
        "/selftools/tools/model/model_export_custom_default_value_condition.html",
        extra_vars={},
    )


@selftools_htmx.route("/selftools/model-import", methods=["POST"])
def selftools_model_import() -> Any | str:
    context: types.Context = cast(
        types.Context,
        {
            "model": model,
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        },
    )
    try:
        tk.check_access("sysadmin", context)
    except tk.NotAuthorized:
        tk.abort(404)

    file = request.files.get("file")

    try:
        data_dict = logic.clean_dict(
            dict_fns.unflatten(logic.tuplize_dict(logic.parse_params(request.form)))
        )
    except dict_fns.DataError:
        return tk.base.abort(400, _("Integrity Error"))

    if not file:
        return "No file provided"

    filename = file.filename if file.filename else ""
    try:
        sfilename = secure_filename(filename)
        f_extension = sfilename.split(".")[-1]
        data = file.read()

        key = (
            data_dict["decryption_key"]
            if data_dict.get("decryption_key")
            else selftools_config.selftools_get_model_ecryption_key()
        )

        if f_extension == "bin" and key:
            fernet = Fernet(key)
            decrypted = fernet.decrypt(data)
            decompressed = gzip.decompress(decrypted).decode("utf-8")
        else:
            decompressed = gzip.decompress(data).decode("utf-8")

        data_dict["models_data"] = json.loads(decompressed)

        resp = tk.get_action("selftools_model_import")(context, data_dict)

        if not resp.get("success"):
            return (
                resp["message"] if resp.get("message") else _("Something went wrong...")
            )
        else:
            return _("Finished.")
    except InvalidToken:
        return _("Cannot decrypt the file.")
    except Exception as e:
        return str(e)



from __future__ import annotations
import logging
from datetime import datetime
from typing import Any
import pytest
import os

from ckan import model
import ckan.plugins.toolkit as tk
from ckan.tests.helpers import call_action
from ckan.tests import factories

from ckanext.selfinfo import config

log = logging.getLogger(__name__)

current_path: list[str] = os.getcwd().split("/")
current_path.pop()
updated_path: str = "/".join(current_path)


@pytest.mark.ckan_config("ckan.plugins", "selfinfo")
@pytest.mark.usefixtures("with_plugins", "clean_db")
@pytest.mark.ckan_config("ckan.selfinfo.ckan_repos_path", updated_path)
@pytest.mark.ckan_config("ckan.selfinfo.ckan_repos", "ckan ckanext-selfinfo")
class TestUPDATE:
    def test_update_last_module_check(self):
        user = factories.User()
        sysadmin = factories.Sysadmin()
        context: dict[str, Any] = {
            "model": model,
            "user": user["name"],
            "ignore_auth": False,
        }

        with pytest.raises(tk.NotAuthorized):
            call_action(config.selfinfo_get_main_action_name(), context=context)

        context["user"] = sysadmin["name"]
        selfinfo: dict[str, Any] = tk.get_action(
            config.selfinfo_get_main_action_name()
        )(context, {})
        assert "python_modules" in selfinfo, selfinfo.keys()
        assert "ckan" in selfinfo["python_modules"], selfinfo["python_modules"].keys()
        assert "ckan" in selfinfo["python_modules"]["ckan"], selfinfo["python_modules"][
            "ckan"
        ].keys()

        ckan_info = selfinfo["python_modules"]["ckan"]["ckan"]

        updated: dict[str, Any] = tk.get_action("update_last_module_check")(
            context, {"module": "ckan"}
        )

        assert ckan_info != updated
        before_updated = datetime.fromisoformat(ckan_info["updated"])
        after_updated = datetime.fromisoformat(updated["updated"])
        assert before_updated < after_updated
        # pop update field, see that they are the same
        ckan_info.pop("updated", None)
        updated.pop("updated", None)
        assert ckan_info == updated



from __future__ import annotations

from typing import Any

from . import get
from . import update
from . import delete
from ckanext.selfinfo import config


def get_actions():
    actions: dict[str, Any] = {
        config.selfinfo_get_main_action_name(): get.get_selfinfo,
        "update_last_module_check": update.update_last_module_check,
        "selfinfo_get_ram": get.selfinfo_get_ram,
        "selfinfo_delete_profile": delete.selfinfo_delete_profile,
    }

    return actions



import click

from . import main


def get_commands():
    return [selfinfo]


@click.group()
def selfinfo():
    """ckanext-selfinfo management commands."""


selfinfo.command(main.update_module_info)
selfinfo.command(main.get_selfinfo)
selfinfo.command(main.write_selfinfo)
selfinfo.command(main.write_selfinfo_duplicated_env)
selfinfo.command(main.delete_selfinfo_redis_key)



from __future__ import annotations

import logging
import json

from ckan.lib.redis import connect_to_redis, Redis
import ckan.plugins.toolkit as tk

from .config import selfinfo_get_errors_limit
from .utils import get_redis_key


class SelfinfoErrorHandler(logging.Handler):
    """Custom handler to store exceptions."""

    def emit(self, record: logging.LogRecord):
        if record.levelno >= logging.ERROR:
            redis: Redis = connect_to_redis()
            log_message = self.format(record)
            redis_key = get_redis_key("errors")

            if not redis.exists(redis_key):
                redis.set(redis_key, json.dumps([]))

            errors = json.loads(redis.get(redis_key))  # pyright: ignore
            errors_limit = selfinfo_get_errors_limit()
            if len(errors) >= errors_limit:
                start_key = len(errors) - errors_limit + 1
                errors = errors[start_key:]
            current_url = None
            try:
                current_url = tk.h.full_current_url()
            except (AttributeError, RuntimeError):
                pass

            errors.append(
                {
                    "error": log_message,
                    "error_url": current_url if current_url else "Missing URL",
                }
            )
            redis.set(redis_key, json.dumps(errors))



from __future__ import annotations

import requests
import logging
import json
from sqlalchemy import desc, exc as sql_exceptions
from sqlalchemy.inspection import inspect
import redis
from typing import Any, Literal
from datetime import datetime

from ckan import types
import ckan.model as model
import ckan.plugins.toolkit as tk
from ckan.lib.search.common import (
    is_available as solr_available,
    make_connection as solr_connection,
)
from ckan.lib.search import clear, rebuild, commit
from ckan.lib.redis import connect_to_redis, Redis

from ckanext.selftools import utils, config

log = logging.getLogger(__name__)


def selftools_solr_query(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any] | Literal[False]:
    tk.check_access("sysadmin", context, data_dict)

    if solr_available():
        solr = solr_connection()
        solr_url = solr.url
        max_limit = config.selftools_get_operations_limit()
        default_search = "q=*:*&rows=" + str(max_limit)

        search = data_dict.get("q", default_search)

        if "rows=" not in search:
            search += "&rows=" + str(max_limit)
        q_response = requests.get(solr_url.rstrip("/") + "/query?" + search)
        q_response.raise_for_status()

        query = q_response.json()

        return query
    return False


def selftools_solr_delete(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any]:
    tk.check_access("sysadmin", context, data_dict)

    if not utils.selftools_verify_operations_pwd(data_dict.get("selftools_pwd")):
        return {"success": False, "message": "Unauthorized action."}

    # Really need to check? It can be not only Datasets
    # pkg = model.Package.get(data_dict.get("id"))
    # if not pkg:
    #     return {"success": False}

    clear(data_dict.get("id", ""))
    return {"success": True}


def selftools_solr_index(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any]:
    tk.check_access("sysadmin", context, data_dict)
    id = data_dict.get("id")
    ids = data_dict.get("ids")

    if not id and not ids:
        return {
            "success": False,
            "message": "Dataset ID or multiple IDs should be provided.",
        }
    pkg = None
    if id:
        pkg = model.Package.get(id)

    try:
        rebuild(
            package_id=pkg.id if pkg else None,
            force=tk.asbool(data_dict.get("force", "False")),
            package_ids=json.loads(ids) if ids else [],
        )
        commit()
    except Exception:
        return {"success": False, "message": "An Error appeared while indexing."}
    return {"success": True}


def selftools_db_query(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any] | Literal[False]:
    tk.check_access("sysadmin", context, data_dict)

    q_model = data_dict.get("model")
    limit = data_dict.get("limit")
    field = data_dict.get("field")
    value = data_dict.get("value")
    order = data_dict.get("order")
    order_by = data_dict.get("order_by")
    if q_model:
        model_fields_blacklist = [
            b.strip().split(".") for b in config.selftools_get_model_fields_blacklist()
        ]
        combained_blacklist = [
            *model_fields_blacklist,
            *[["User", "password"], ["User", "apikey"]],
        ]

        def _get_db_row_values(row: Any, columns: Any, model_name: str) -> list[Any]:
            values = []
            for col in columns:
                if [
                    b for b in combained_blacklist if b[0] == model_name and col == b[1]
                ]:
                    value = "SECURE"
                else:
                    value = getattr(row, col, None)

                if value is not None:
                    values.append(value)
                else:
                    values.append("")

            return values

        models = utils.get_db_models()
        curr_model = [m for m in models if m["label"] == q_model]

        if curr_model:
            try:
                model_class = curr_model[0]["model"]
                q = model.Session.query(model_class)

                if field and value:
                    q = q.filter(getattr(model_class, field) == value)

                if order_by and order:
                    if order == "desc":
                        q = q.order_by(desc(order_by))
                    else:
                        q = q.order_by(order_by)

                if limit:
                    q = q.limit(int(limit))

                results = q.all()

                columns = [col.name for col in inspect(model_class).c]

                structured_results = [
                    _get_db_row_values(row, columns, curr_model[0]["label"])
                    for row in results
                ]

                return {
                    "success": True,
                    "results": structured_results,
                    "fields": columns,
                }
            except (AttributeError, sql_exceptions.CompileError) as e:
                return {
                    "success": False,
                    "message": str(e),
                }
    return False


def selftools_db_update(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any]:
    tk.check_access("sysadmin", context, data_dict)

    if not utils.selftools_verify_operations_pwd(data_dict.get("selftools_pwd")):
        return {"success": False, "message": "Unauthorized action."}

    q_model = data_dict.get("model")
    limit = data_dict.get("limit")
    field = data_dict.get("field")
    value = data_dict.get("value")
    where_field = data_dict.get("where_field")
    where_value = data_dict.get("where_value")
    if q_model:
        models = utils.get_db_models()
        curr_model = [m for m in models if m["label"] == q_model]

        if curr_model:
            try:
                model_class = curr_model[0]["model"]
                table_details = inspect(model_class)

                primary_key = None
                try:
                    primary_name = table_details.primary_key[0].name
                    primary_key = getattr(model_class, primary_name)
                except Exception:
                    return {
                        "success": False,
                        "message": "Cannot extract Primary key for the Model.",
                    }

                # First filter and limit results
                q = model.Session.query(primary_key)

                if where_field and where_value:
                    q = q.filter(getattr(model_class, where_field) == where_value)

                if limit:
                    q = q.limit(int(limit))

                if field and value:
                    ids = [i[0] for i in q.all()]
                    # Update already limited results
                    upd = (
                        model.Session.query(model_class)
                        .filter(primary_key.in_(ids))
                        .update({field: value})
                    )

                    model.Session.commit()

                    return {
                        "success": True,
                        "updated": upd,
                        "effected": ids,
                        "effected_json": json.dumps(ids, indent=2),
                    }
                else:
                    return {"success": False, "message": "Provide the WHERE condition"}
            except AttributeError:
                return {
                    "success": False,
                    "message": f"There no attribute '{field}' in '{curr_model[0]['label']}'",
                }

    return {"success": False}


def selftools_redis_query(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any] | Literal[False]:
    tk.check_access("sysadmin", context, data_dict)

    def _redis_key_value(redis_conn: Any, key: str):
        key_type = redis_conn.type(key).decode("utf-8")
        val = ""
        try:
            if key_type == "string":
                val = redis_conn.get(key)
            elif key_type == "hash":
                val = redis_conn.hgetall(key)
            else:
                val = f"<Unsupported type: {key_type}>"
        except redis.exceptions.RedisError as e:  # pyright: ignore
            val = f"<Error: {str(e)}>"

        return val

    def _safe_key_display(k: bytes) -> str:
        try:
            # Check for binary prefix or signs of pickled data
            if any(s in repr(k) for s in [r"\x80", r"\x00"]):
                return repr(k)
            return k.decode("utf-8")
        except UnicodeDecodeError:
            return repr(k)

    redis_conn: Redis = connect_to_redis()

    q = data_dict.get("q", "")
    if q:
        keys = redis_conn.keys(f"*{q}*")
        max_limit = config.selftools_get_operations_limit()
        keys = keys[:max_limit]  # pyright: ignore
        redis_results = [
            {
                "key": _safe_key_display(k),
                "type": redis_conn.type(k).decode("utf-8"),  # pyright: ignore
                "value": str(_redis_key_value(redis_conn, k)),
            }
            for k in keys
        ]

        return {"success": True, "results": redis_results}
    return False


def selftools_redis_update(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any]:
    tk.check_access("sysadmin", context, data_dict)

    if not utils.selftools_verify_operations_pwd(data_dict.get("selftools_pwd")):
        return {"success": False, "message": "Unauthorized action."}

    key = data_dict.get("redis_key")
    value = data_dict.get("value")
    if key and value:
        redis_conn: Redis = connect_to_redis()
        redis_conn.set(key, value)
        return {"success": True}

    return {"success": False}


def selftools_redis_delete(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any]:
    tk.check_access("sysadmin", context, data_dict)

    if not utils.selftools_verify_operations_pwd(data_dict.get("selftools_pwd")):
        return {"success": False, "message": "Unauthorized action."}

    key = data_dict.get("redis_key")
    if key:
        redis_conn: Redis = connect_to_redis()

        deleted = redis_conn.delete(key)
        if deleted:
            return {"success": True}
    return {"success": False}


def selftools_config_query(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any]:
    tk.check_access("sysadmin", context, data_dict)

    if not utils.selftools_verify_operations_pwd(data_dict.get("selftools_pwd")):
        return {"success": False, "message": "Unauthorized action."}

    key = data_dict.get("q")
    if key:
        blacklist = config.selftools_get_config_blacklist()
        default_blacklist = [
            "sqlalchemy.url",
            "ckan.datastore.write_url",
            "ckan.datastore.read_url",
            "solr_url",
            "solr_user",
            "solr_password",
            "ckan.redis.url",
            config.SELFTOOLS_CONFIG_BLACKLIST,
            config.SELFTOOLS_OPERATIONS_PWD,
        ]
        config_keys = tk.config.keys()
        config_keys = [
            k for k in config_keys if k not in [*default_blacklist, *blacklist]
        ]
        config_results = [
            {"key": ck, "value": tk.config.get(ck)} for ck in config_keys if key in ck
        ]
        return {"success": True, "results": config_results}

    return {"success": False}


def selftools_model_export(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any] | Literal[False]:
    tk.check_access("sysadmin", context, data_dict)

    if not utils.selftools_verify_operations_pwd(data_dict.get("selftools_pwd")):
        return {"success": False, "message": "Unauthorized action."}

    q_model = data_dict.get("model")
    limit = data_dict.get("limit")
    locals = data_dict.get("local[]", [])
    remotes = data_dict.get("remote[]", [])
    custom_relationships = []

    if locals and remotes:
        for r in (
            list(zip(locals, remotes))
            if isinstance(locals, list)
            else [(locals, remotes)]
        ):
            local = r[0].split(".")
            remote = r[1].split(".")

            if not len(local) == 2 or not len(remote) == 2:
                return {
                    "success": False,
                    "message": "Issue with extracting Custom Relationships, please review them.",
                }
            custom_relationships.append(
                {
                    "local_model": local[0],
                    "local_field": local[1],
                    "remote_model": remote[0],
                    "remote_field": remote[1],
                }
            )

    # default value condition zip
    default_value_conditions_field = [
        "default_value_condition_model[]",
        "default_value_condition_field[]",
        "default_value_condition_value[]",
        "default_value_condition_set_field[]",
        "default_value_condition_set_value[]",
    ]
    columns = map(lambda key: data_dict.get(key, []), default_value_conditions_field)
    default_value_conditions = list(zip(*columns))

    if q_model:

        def _to_string(value: str):
            if isinstance(value, datetime):
                return value.isoformat()
            return value

        def _collect(
            row: dict[str, Any],
            model_class: Any,
            collector: dict[str, Any],
            default_models: Any,
        ) -> None | dict[str, Any]:
            model_name = model_class.__name__
            table_details = inspect(model_class)

            primary_name = None
            try:
                primary_name = table_details.primary_key[0].name
            except Exception:
                return {
                    "success": False,
                    "message": "Cannot extract Primary name for the Model.",
                }

            relationships = table_details.relationships
            columns = [col.name for col in table_details.c]

            values = {}
            primary_value = ""
            for col in columns:
                value = getattr(row, col, None)

                if value is not None:
                    values[col] = _to_string(value)
                else:
                    values[col] = None

                if col == primary_name:
                    primary_value = value

            unique_key = model_name + "." + primary_value

            if unique_key not in collector:
                collector[unique_key] = {
                    "model": model_name,
                    "primary_key": primary_name,
                    "values": values,
                }

                for _, rel in relationships.items():
                    class_field = None
                    local_key = None
                    for local_col, remote_col in rel.local_remote_pairs:
                        remote_key = remote_col.name
                        local_key = local_col.name

                        class_field = getattr(rel.mapper.class_, remote_key)
                    if class_field and local_key:
                        rows = (
                            model.Session.query(rel.mapper.class_)
                            .filter(class_field == values[local_key])
                            .all()
                        )
                        if rows:
                            [
                                _collect(
                                    row, rel.mapper.class_, collector, default_models
                                )
                                for row in rows
                            ]

            if c_r := [
                i for i in custom_relationships if i["local_model"] == model_name
            ]:
                for r in c_r:
                    r_m = [m for m in default_models if m["label"] == r["remote_model"]]
                    if r_m and r["local_field"] in values:
                        remote_model = r_m[0]["model"]
                        class_field = getattr(remote_model, r["remote_field"])

                        rows = (
                            model.Session.query(remote_model)
                            .filter(class_field == values[r["local_field"]])
                            .all()
                        )
                        if rows:
                            [
                                _collect(row, remote_model, collector, default_models)
                                for row in rows
                            ]

            if default_value_conditions:
                matches = [
                    condition
                    for condition in default_value_conditions
                    if model_name == condition[0]
                    and condition[1] in values
                    and condition[2] == values[condition[1]]
                ]

                if matches:
                    for m in matches:
                        collector[unique_key]["values"][m[3]] = m[4]

        default_models = utils.get_db_models()
        curr_model = [m for m in default_models if m["label"] == q_model]

        if curr_model:
            field = data_dict.get("field")
            value = data_dict.get("value")

            try:
                model_class = curr_model[0]["model"]
                q = model.Session.query(model_class)

                if field and value:
                    q = q.filter(getattr(model_class, field) == value)

                if limit:
                    q = q.limit(int(limit))

                results = q.all()
                collector = {}
                [
                    _collect(row, model_class, collector, default_models)
                    for row in results
                ]

                return {
                    "success": True,
                    "results": collector,
                }
            except (AttributeError, sql_exceptions.CompileError) as e:
                return {
                    "success": False,
                    "message": str(e),
                }
    return False


def selftools_model_import(
    context: types.Context, data_dict: dict[str, Any]
) -> dict[str, Any] | Literal[False]:

    tk.check_access("sysadmin", context, data_dict)

    if not utils.selftools_verify_operations_pwd(data_dict.get("selftools_pwd")):
        return {"success": False, "message": "Unauthorized action."}

    default_models = utils.get_db_models()
    inserted_list = []
    models_data = data_dict.get("models_data", {})

    def _get_model_class(name: str):
        return [m for m in default_models if m["label"] == name][0]["model"]

    def _get_model_relations(model_names: Any) -> dict[str, Any]:
        """
        Returns:
        {
            'PackageExtra': {
                'Package': {
                    'local_column': 'package_id',
                    'remote_column': 'id'
                }
            },
            ...
        }
        """
        dependencies = {}

        for model_name in model_names:
            model_cls = _get_model_class(model_name)
            deps = {}

            for rel in inspect(model_cls).relationships:
                if not rel.direction.name.startswith("MANYTOONE"):
                    continue  # Only track dependencies, not backrefs

                related_cls = rel.mapper.class_
                related_model_name = related_cls.__name__

                if related_model_name in model_names:
                    # Should be a 1:1 mapping between local and remote columns
                    local_column = next(iter(rel.local_columns)).name
                    remote_column = next(iter(rel.remote_side)).name

                    deps[related_model_name] = {
                        "local_column": local_column,
                        "remote_column": remote_column,
                    }

            dependencies[model_name] = deps

        return dependencies

    def _try_to_datetime(value: Any):
        if value and isinstance(value, str):
            try:
                return datetime.fromisoformat(value)
            except ValueError:
                pass

        return value

    def _create_row(rid: str, row: dict[str, Any]):
        if rid in inserted_list:
            return

        model_name = row["model"]
        model_class = _get_model_class(model_name)
        values = row["values"]
        primary_name = row["primary_key"]

        if model_relations.get(model_name):
            # Insert firstly the parent related row
            for m, k in model_relations[model_name].items():
                local_value = values.get(k["local_column"])

                dict_key = ".".join([m, local_value])

                if models_data.get(dict_key):
                    _create_row(dict_key, models_data[dict_key])

        primary_key = getattr(model_class, primary_name)

        session = model.Session()

        obj_id = values.get(primary_name)
        r_obj = model.Session.query(model_class).filter(primary_key == obj_id).first()

        if not r_obj:
            r_obj = model_class()
        for k, v in values.items():
            setattr(r_obj, k, _try_to_datetime(v))

        try:
            session.add(r_obj)
            session.commit()
            inserted_list.append(rid)
        except sql_exceptions.IntegrityError:
            session.rollback()

    try:
        model_classes = {v["model"] for v in models_data.values()}
        model_relations = _get_model_relations(model_classes)

        for rid, row in models_data.items():
            _create_row(rid, row)
    except Exception as e:
        return {
            "success": False,
            "message": str(e),
        }
    return {"success": True}



from __future__ import annotations

import logging

from typing import Any
import ckan.model as model
import ckan.plugins as p

from ckanext.selftools import interfaces
from ckanext.selftools.config import (
    selftools_get_operations_pwd,
    selftools_get_categories_list,
    selftools_get_tools_blacklist,
)

log = logging.getLogger(__name__)

SELFTOOLS_TOOLS = [
    {
        "key": "solr",
        "label": "Solr",
        "tools": [
            {
                "key": "solr_query",
                "label": "Query",
                "snippet": "/selftools/tools/solr/solr_query.html",
            },
            {
                "key": "solr_index",
                "label": "Index",
                "snippet": "/selftools/tools/solr/solr_index.html",
            },
            {
                "key": "solr_delete",
                "label": "Delete",
                "snippet": "/selftools/tools/solr/solr_delete.html",
            },
        ],
    },
    {
        "key": "db",
        "label": "DB",
        "tools": [
            {
                "key": "db_query",
                "label": "Query",
                "snippet": "/selftools/tools/db/db_query.html",
            },
            {
                "key": "db_update",
                "label": "Update",
                "snippet": "/selftools/tools/db/db_update.html",
            },
        ],
    },
    {
        "key": "redis",
        "label": "Redis",
        "tools": [
            {
                "key": "redis_query",
                "label": "Query",
                "snippet": "/selftools/tools/redis/redis_query.html",
            },
            {
                "key": "redis_update",
                "label": "Update/Create",
                "snippet": "/selftools/tools/redis/redis_update.html",
            },
            {
                "key": "redis_delete",
                "label": "Delete",
                "snippet": "/selftools/tools/redis/redis_delete.html",
            },
        ],
    },
    {
        "key": "config",
        "label": "Config",
        "tools": [
            {
                "key": "config_query",
                "label": "Query",
                "snippet": "/selftools/tools/config/config_query.html",
            },
        ],
    },
    {
        "key": "model",
        "label": "Model",
        "tools": [
            {
                "key": "model_export",
                "label": "Export",
                "snippet": "/selftools/tools/model/model_export.html",
            },
            {
                "key": "model_import",
                "label": "Import",
                "snippet": "/selftools/tools/model/model_import.html",
            },
        ],
    },
]


def get_db_models() -> list[dict[str, Any]]:
    try:
        models = [
            model.Package,
            model.PackageExtra,
            model.PackageTag,
            model.PackageRelationship,
            model.Tag,
            model.Resource,
            model.ResourceView,
            model.User,
            model.Group,
            model.GroupExtra,
            model.Member,
            model.PackageMember,
            model.Vocabulary,
            model.SystemInfo,
            model.ApiToken,
        ]

        # models modification
        for item in p.PluginImplementations(interfaces.ISelftools):
            item.selftools_db_models(models)

        return [{"label": model.__name__, "model": model} for model in models]
    except Exception:
        log.error("Cannot retrieve DB Models.")

    return [{}]


def get_selftools_categories() -> list[dict[str, Any]]:
    tools_blacklist = selftools_get_tools_blacklist()

    def _filter_tools(category: dict[str, Any]) -> dict[str, Any]:
        tools = category.get("tools")
        if tools_blacklist and tools:
            for tb in tools_blacklist:
                tb = tb.strip().split(".")
                if category["key"] == tb[0]:
                    tools = [t for t in tools if t["key"] != tb[1]]
            category["tools"] = tools
        return category

    categories = [
        _filter_tools(c)
        for c in SELFTOOLS_TOOLS
        if c["key"] in selftools_get_categories_list()
    ]

    return categories


def selftools_verify_operations_pwd(pwd: str | None) -> bool:
    config_pwd = selftools_get_operations_pwd()
    if not config_pwd:
        return True

    if config_pwd and pwd and (config_pwd == pwd):
        return True

    return False



from __future__ import annotations

import logging
from typing import Any

from ckanext.selfinfo import utils
import ckan.plugins.toolkit as tk

from ckan import types

log = logging.getLogger(__name__)


def selfinfo_delete_profile(context: types.Context, data_dict: dict[str, Any]) -> bool:
    tk.check_access("sysadmin", context, data_dict)
    return utils.selfinfo_delete_redis_key(data_dict.get("profile", ""))



import pytest

from ckan.plugins import plugin_loaded

"""
Tests for plugin.py.

Tests are written using the pytest library (https://docs.pytest.org), and you
should read the testing guidelines in the CKAN docs:
https://docs.ckan.org/en/2.9/contributing/testing.html

To write tests for your extension you should install the pytest-ckan package:

    pip install pytest-ckan

This will allow you to use CKAN specific fixtures on your tests.

For instance, if your test involves database access you can use `clean_db` to
reset the database:

    import pytest

    from ckan.tests import factories

    @pytest.mark.usefixtures("clean_db")
    def test_some_action():

        dataset = factories.Dataset()

        # ...

For functional tests that involve requests to the application, you can use the
`app` fixture:

    from ckan.plugins import toolkit

    def test_some_endpoint(app):

        url = toolkit.url_for('myblueprint.some_endpoint')

        response = app.get(url)

        assert response.status_code == 200


To temporary patch the CKAN configuration for the duration of a test you can use:

    import pytest

    @pytest.mark.ckan_config("ckanext.myext.some_key", "some_value")
    def test_some_action():
        pass
"""


@pytest.mark.ckan_config("ckan.plugins", "selfinfo")
@pytest.mark.usefixtures("with_plugins")
def test_plugin():
    assert plugin_loaded("selfinfo")



import ckan.plugins as plugins
import ckan.plugins.toolkit as tk
from ckan.common import CKANConfig

from .logic import action
from . import cli


@tk.blanket.config_declarations
@tk.blanket.cli(cli.get_commands)
@tk.blanket.actions(action.get_actions)
@tk.blanket.blueprints
@tk.blanket.helpers
class SelfinfoPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)

    # IConfigurer

    def update_config(self, config_: CKANConfig):
        tk.add_template_directory(config_, "templates")
        tk.add_public_directory(config_, "public")
        tk.add_resource("assets", "selfinfo")
